import sympy
from sympy import diff, oo
from .sosconstr import sosconstr


def preprocess(symexpr, var, info):
    # Get the maximum degree of the independent variable
    maxdeg = 0
    dummy = diff(symexpr, var)
    while dummy != 0:
        maxdeg += 1
        dummy = diff(dummy, var)

    # Substitute var
    if info[1] == oo:
        newvar = var + info[0]
        newsymexpr = symexpr.subs(var, newvar)
    elif info[0] == -oo:
        newvar = -var + info[1]
        newsymexpr = symexpr.subs(var, newvar)
    else:
        newvar = (info[1] - info[0]) / 2 * (1 - var) / (1 + var) + (info[1] + info[0]) / 2
        newsymexpr = symexpr.subs(var, newvar) * (1 + var) ** maxdeg

    return newsymexpr



def sosineq(sos, symexpr, info1=None, info2=None):
    '''
    SOSINEQ   Adds an inequality constraint to the SOS program.
    SOSP = sosineq(SOSP,EXPR)
    
    SOSP is the sum of squares program.
    The new constraint is described by EXPR as follows:
    EXPR is a symbolic expression in the variables of the program.
    '''

    # Check if symexpr is a Matrix
    scal_expr = 0
    if not isinstance(symexpr,sympy.MatrixBase):
        scal_expr = 1

    # Interval
    if info1 is not None and not isinstance(info1, str):
        if len(sos.symvartable) != 1:
            raise ValueError('Not a univariate polynomial.')
        if scal_expr != 1:
            for i in range(symexpr.shape[0]):
                #print(symexpr[i])   # 8/21: Why only one index? A vector?
                symexpr[i] = preprocess(symexpr[i], sos.symvartable[0], info1) ## It seems symexpr is a vector?
                sos = sosconstr(sos, 'ineq', symexpr[i])
                sos.expr['type'][sos.expr['num']-1] = 'posint'  # overwrite the type
        else:
            symexpr = preprocess(symexpr, sos.symvartable[0], info1)
            sos = sosconstr(sos, 'ineq', symexpr)
            sos.expr['type'][sos.expr['num']-1] = 'posint'  # overwrite the type

        return sos
    
    # Sparse polynomial
    if info1 == 'sparse':
        if scal_expr != 1:
            for i in range(symexpr.shape[0]):
                sos = sosconstr(sos, 'ineq', symexpr[i])
                sos.expr['type'][sos.expr['num']-1] = 'sparse'  # overwrite the type
        else:
            sos = sosconstr(sos, 'ineq', symexpr)
            sos.expr['type'][sos.expr['num']-1] = 'sparse'  # overwrite the type

        return sos
    

    # Sparse polynomial multipartite
    if info1 == 'sparsemultipartite':
        if scal_expr != 1:
            for i in range(symexpr.shape[0]):
                sos = sosconstr(sos, 'ineq', symexpr[i])
                sos.expr['type'][sos.expr['num']-1] = 'sparsemultipartite'
                sos.expr['multipart'][sos.expr['num']-1] = info2  # 8/18 Edited: is it overwriting? like the above?
        else:
            sos = sosconstr(sos, 'ineq', symexpr)
            sos.expr['type'][sos.expr['num']-1] = 'sparsemultipartite'
            sos.expr['multipart'][sos.expr['num']-1] = info2  # 8/18 Edited: is it overwriting? like the above?

        return sos


    # If info1 and infor2 are empty, then the inequality constraint is
    sos = sosconstr(sos, 'ineq', symexpr)
    return sos


import numpy as np
from sympy import poly

def getdegrees(Z, symvartable):
    '''
    GETDEGREES --- Get degrees of monomials.

    ZDEG = getdegrees(Z,VARTABLE)

    Z is a list of monomials.
    VARTABLE is the list of independent variables.
    ZDEG is the degree of Z (in SOSTOOLS notation)
    '''

    p = np.sum(Z)
    p = poly(p, symvartable)
    coefmonmatr = [[*monomial] for monomial in p.monoms()]
    Zdeg = np.array(coefmonmatr)

    return  Zdeg


from sympy import Matrix
from .sosconstr import sosconstr

def soseq(sos, symexpr):
    '''
    sos = soseq(sos,symexpr)
    SOSEQ --- Add a new equality constraint f(x) = 0
    to an SOS program 

    OUTPUTS:
    SOSP: The modified SOS program structure.

    INPUTS:
    SOSP: The SOS program structure to be modified
    SYMEXPR: The expression on the left hand side of the constraint, i.e., f(x).
    This expression may be scalar, matrix, or vector-valued. 
    '''

    # Check if symexpr is a Matrix
    if not isinstance(symexpr, Matrix):
        sos = sosconstr(sos, 'eq', symexpr)
    else:
        dimp = symexpr.shape[0]
        for i in range(dimp):
            sos = sosconstr(sos, 'eq', symexpr[i])

    return sos
    
        


from .getequation import getequation

def sosconstr(sos,Type,symexpr):
    '''
    SOSCONSTR   Adds a constraint to the SOS program.
    SOSP = sosconstr(SOSP,TYPE,EXPR)
    
    SOSP is the sum of squares program.
    The new constraint is described by TYPE as follows:
    '''
    expr = sos.expr['num']     # expr is an index
    sos.expr['num'] = sos.expr['num'] + 1
    sos.expr['type'][expr] = Type  # Update the type of the constraint

    # assume symexpr is a sympy expression, and sos.symvartable, sos.symdecvartable and sos.varmat.symvartable are lists of sympy expressions.
    sos.expr['At'][expr], sos.expr['b'][expr], sos.expr['Z'][expr] = getequation(symexpr, sos.symvartable, sos.symdecvartable, sos.varmat['symvartable'])

    return sos


from sympy import expand, Matrix
from .sosprogram import sosprogram
from .sosineq import sosineq
from .sossolve import sossolve
import numpy as np
import math
from scipy.linalg import sqrtm
from .proj3 import proj3


def findsos(P,flag='abcdefgh',options=None):
    '''
    FINDMATRIXSOS --- Find a sum of squares decomposition of a given matrix polynomial.

    [Q,Z,decomp,Den] = findsos(P,flag,options)

    P is a symmetric polynomial matrix.

    FLAG is an optional argument which, if set to 'rational', returns a
    sum of squares decomposition with rational coefficients. 

    OPTIONS is an optional argument which specifies the solver and the
    solver-specific parameters on its fields as
    options.solver
    options.params
    the default value for options.solver is 'Mosek'. The solver parameters
    are fields of options.params as, for instance, options.params.tol = 1e-9.

    A positive semidefinite Q and a symbolic monomial vector Z will be
    computed such that

    (Ir kron Z)' * Q * (Ir kron Z) = P(x)

    If P is not a sum of squares, the function will return empty Q and Z.

    If P is a polynomial with integer coefficients and is represented as a
    symbolic object, then [Q,Z] = findsos(P,'rational') will compute a rational
    matrix Q such that Z'*Q*Z = P.
    '''

    if isinstance(P,Matrix):
        dimp = P.shape
        if P.shape[0] != P.shape[1]:
            raise ValueError('The polynomial matrix is not square, it cannot be a sum of squares.')
        if not P.is_symmetric():
            raise ValueError('The polynomial matrix is not symmetric, it can not be a sum of squares.')
    else:
        dimp = [1,1]
    

    P = expand(P)   # Do we need this?

    vartable = list(P.free_symbols)

    # Initialize the SOS program
    prog = sosprogram(vartable)

    # Add the constraint
    prog = sosineq(prog, P)

    # Solve the SOS program
    if options is None:
        options = {}
        options['solver'] = 'mosek'
        # params = {}
        # params['tol'] = 1e-9
        # params['alg'] = 2
        # options['params'] = params
    
    prog = sossolve(prog, options)
    info = prog.solinfo['info']
    
    if 'solver' in options and options['solver'] == 'SDPNAL':
        print('findsos function currently not supported for SDPNAL solver')
        return [],[],[],[]
    elif info['dinf']==1 or info['pinf']==1:
        print('No sum of squares decomposition is found.')
        return [],[],[],[]
    else:
        len_RRx = len(prog.solinfo['RRx'])
        Q = np.array(prog.solinfo['RRx']).reshape(int(math.sqrt(len_RRx)), int(math.sqrt(len_RRx)))
        Z = []
        Z_temp = prog.extravar['Z'][0].toarray()
        for row in Z_temp:
            Z.append(np.prod([var**deg for var, deg in zip(vartable, row)]))
    
    if flag == 'rational':
        A = prog.expr['At'][0].toarray().T
        B = prog.expr['b'][0].toarray()

        # Round x0
        N = 1   # This is the tolerance in the rounding
        xx = prog.solinfo['x']

        kmax = 10
        pd = 0
        while kmax != 0:
            kmax = kmax - 1
            x0 = np.round(N*xx).reshape(-1,1)
            try:
                Qflat,NN = proj3(x0,A,B,N)
                n = math.sqrt(len(Qflat))
                Qr = Qflat.reshape(int(n),int(n)).T
                # Qr should be PSD
                if np.min(np.linalg.eig(Qr/NN)[0]) > -1e-14:
                    kmax = 0
                    pd = 1
                # Increase N, and try again
                N = 2*N
            except:
                NN = 1
                n = math.sqrt(len(xx))
                Qr = xx.reshape(n,n).T
                break
        
        # Experimental, no good error checking yet, so we check that expand(NN*P - Z.'*Qr*Z) is zero!
        try:
            check = expand(NN*P - Matrix(Z).transpose() @ Matrix(Qr) @ Matrix(Z))
            if check != 0 or pd == 0:
                Qr = []
                Z = []
                NN = []
                raise Exception('Could not compute a rational SOS!')
        except:
            Qr = []
            Z = []
            NN = []
            raise Exception('Could not compute a rational SOS!')
        
        Q = Qr
        Den = NN

    else:
        Den = 1

    L = np.real(sqrtm(Q))
    decomp = L @ np.kron(np.eye(dimp[0]), np.array(Z).reshape(-1,1))

    return Q, Z, decomp, Den

        
    






from .inconvhull import inconvhull
import numpy as np

def sparsemultipart(Z1,Z2,info):
    '''
    Find the elements in Z1 that are in the convex hull of Z2, where Z2 is bipartite
    '''
    
    I = [i for i,x in enumerate(Z2.sum(axis=0).tolist()[0]) if x==0]
    sizeinf = len(info)

    if sizeinf == 1:
        raise Exception('Error in sparsemultipart option - at least two sets are required')
    Z3 = []
    for i in range(sizeinf):
        Z3_temp = inconvhull(Z1[:,info[i]],Z2[:,info[i]])
        Z3.append(Z3_temp)
        
    for i in range(sizeinf-1):
        Z3[i+1] = np.hstack([np.vstack([Z3[i]]*Z3[i+1].shape[0]),np.kron(Z3[i+1],np.ones((Z3[i].shape[0],1)))])
        
    Z3 = Z3[-1]
    Z = np.zeros((Z3.shape[0],Z2.shape[1]))
    lgth = 0

    for i in range(sizeinf):
        Z[:,info[i]] = Z3[:,lgth:(lgth+len(info[i]))]
        lgth = len(info[i])+lgth

    Z[:,I] = np.zeros((Z.shape[0],len(I)))

    return Z


from sympy import Matrix
from .sospolyvar import sospolyvar

def sospolymatrixvar(sos,ZSym,n,matrixstr=None):
    '''
    SOSPOLYMATRIXVAR --- Declare a polynomial matrix variable P in the sos program
    SOS of

    SOSP,P = sospolymatrixvar(SOSP,ZSym,n,matrixstr)

    OUTPUTS:
    SOS: the modified sum of squares program structure.
    P: the new polynomial matrix variable. If SOS contains symbolic variables, the
    output type of P will be sym. Otherwise, the output type defaults to
    dpvar unless the caller specifies output='pvar' as a fifth input.

    INPUTS
    SOS: The sum of squares program structure to be modified.
    ZSym[list]: The list of monomials to be contained in P. Decision
    variables corresponding to those monomials will be assigned
    automatically by SOSPOLYVAR.
    n: The desired dimension of the output matrix P: n(1) x n(2)
    matrixstr (optional): a char string with the option 'symmetric' when required
    '''

    if matrixstr is not None:
        if matrixstr == 'symmetric':
            if n[0] == n[1]:
                P = Matrix.zeros(n[0])
                for i in range(n[0]):
                    for j in range(n[0]):
                        sos,var = sospolyvar(sos,ZSym)
                        P[i,j] = var
                        P[j,i] = var

                return sos, P
            # omit two other cases in MATLAB code
            else:
                print("'symmetric' option used, matrix must be square.")
                P = []
                return sos, P
        else:
            print("matrix structure 'matrixstr' is not defined.")
            P = []
            return sos, P
        
    else:
        
        P = Matrix.zeros(n[0],n[1])
        for i in range(n[0]):
            for j in range(n[1]):
                sos,var = sospolyvar(sos,ZSym)
                P[i,j] = var
        if n[0] == n[1] == 1:
            P = P[0,0]

        return sos, P


        


from .sosprogram import sosprogram
from .findcommonZ import findcommonZ
from .findsos import findsos
from .getconstraint import getconstraint
from .getdegrees import getdegrees
from .getequation import getequation
from .monomials import monomials
from .sosconstr import sosconstr
from .soseq import soseq
from .sosgetsol import sosgetsol
from .sosineq import sosineq
from .sosmatrixineq import sosmatrixineq
from .sospolymatrixvar import sospolymatrixvar
from .sospolyvar import sospolyvar
from .sossetobj import sossetobj
from .sossolve import sossolve
from .sossosvar import sossosvar
from .sosvar import sosvar
from .spantiblkdiag import spantiblkdiag
from .spblkdiag import spblkdiag
from .inconvhull import inconvhull
from .sparsemultipart import sparsemultipart
from .proj3 import proj3


from sympy import symbols
from .sosvar import sosvar

def sospolyvar(sos,Zsym):
    '''
    SOSPOLYVAR --- Declare a new scalar polynomial variable in
        an SOS program

    SOS, V = sospolyvar(sos,ZSym)

    OUTPUTS:
    SOS: the modified sum of squares program structure.
    V: the new polynomial variable.

    INPUTS
    SOS: The sum of squares program structure to be modified.
    ZSym: The vector of monomials to be contained in V. Decision
    variables corresponding to those monomials will be assigned
    automatically by SOSPOLYVAR.
    '''

    if (isinstance(Zsym, int) or isinstance(Zsym, float)) and Zsym == 1:
        Zsym = symbols(Zsym)

    sos, V = sosvar(sos,'poly',Zsym)

    return sos, V



from scipy.sparse import csr_matrix, vstack, hstack, eye, issparse
import numpy as np
import pandas as pd

def findcommonZ(Z1, Z2):
    '''
    FINDCOMMONZ --- Find common(distinct) Z and permutation matrices R1, R2

    R1,R2,Z = findcommonZ(Z1,Z2)

    Given two vectors of monomials Z1 and Z2, this 
    function will compute another vector of monomials Z
    containing all the  distinct monomials of Z1 and Z2, and
    permutation matrices R1, R2 such that

    Z1 = R1*Z
    Z2 = R2*Z

    Assumption: all the monomials in Z1, as well as
    the monomials in Z2, are DISTINCT --- but Z1 and Z2 may 
    have common monomials.
    '''

    # Check if Z1 and Z2 are sparse matrix
    if not issparse(Z1):
        Z1 = csr_matrix(Z1)
    if not issparse(Z2):
        Z2 = csr_matrix(Z2)

    if (Z1.shape[0] + Z2.shape[0]) <= 1:
        Z = vstack([Z1, Z2])
        R1 = eye(Z1.shape[0], Z.shape[0])
        R2 = eye(Z2.shape[0], Z.shape[0])

        return R1, R2, Z
        
    # Constructing index matrix
    sizeZ1 = Z1.shape[0]
    Ind1 = np.arange(sizeZ1)[:, None]
    sizeZ2 = Z2.shape[0]
    Ind2 = np.arange(sizeZ2)[:, None]
    Ind = np.block([[Ind1, np.full(Ind1.shape, sizeZ2)], [np.full(Ind2.shape, sizeZ1), Ind2]])
    
    # Constructing Z
    ZZ = vstack([Z1, Z2])
    ZZ_temp = pd.DataFrame(ZZ.toarray())
    IndSort = ZZ_temp.sort_values(by=list(ZZ_temp.columns)).index
    ZZ = ZZ[IndSort]
    sizeZZ = ZZ.shape[0]
    ZTemp = np.diff(ZZ.toarray(), prepend=ZZ[-1:].toarray(), axis=0)    # Functionally equivalent to MATLAB code
    I = np.where(np.any(ZTemp != 0, axis=1))[0]
    INull = np.where(np.all(ZTemp == 0, axis=1))[0]
    if I.size == 0:
        I = 0
        INull = 1
    Z = ZZ[I]

    # Constructing permutation matrix
    Ind = Ind[IndSort]
    for i in INull:
        Ind[i - 1, 1] = Ind[i, 1]
        Ind[i, 1] = sizeZ2
    Ind = Ind[I]

    # hstack in scipy.sparse
    R1 = hstack([eye(sizeZ1), csr_matrix((sizeZ1, len(I) - sizeZ1))]).tocsr()
    R1 = R1[:, Ind[:, 0]]
    R2 = hstack([eye(sizeZ2), csr_matrix((sizeZ2, len(I) - sizeZ2))]).tocsr()
    R2 = R2[:, Ind[:, 1]]

    Z = csr_matrix(Z)

    return R1, R2, Z


from .getdegrees import getdegrees
from scipy.sparse import csr_matrix, eye, hstack, vstack, find
from .getconstraint import getconstraint
from sympy import symbols, Matrix, prod
import numpy as np

# This function is used when Type == 'sos'
def myAdecvarprod(A,decvar):
    rows, cols, _ = find(A)

    sympoly = Matrix.zeros(A.shape[1], 1)

    for col in range(A.shape[1]):
        idx = [row for row in rows[cols == col]]
        
        if idx:
            sympoly[col, 0] = sum(decvar[i] for i in idx)

    return sympoly


def sosvar(sos,Type,Zsym):
    '''
    SOSVAR --- Declare a new SOSP variable in an SOS program 

    sos, V = sosvar(SOSP,TYPE,Z)

    SOSP is the sum of squares program.
    VAR is the new SOSP variable. This variable is described by TYPE and a 
    vector of monomial Z in the following manner:
    TYPE = 'sos'  : the new SOSP variable is a sum of squares        --- Z' * Q * Z
    TYPE = 'poly' : the new SOSP variable is an ordinary polynomial  --- F * Z
    '''

    if Type != 'sos' and Type != 'poly':
        raise Exception('Unknown type.')
    
    Z = getdegrees(Zsym, sos.symvartable)   # Z is a expononet matrix of monomials

    # Add new variable
    sos.var['num'] += 1
    var = sos.var['num']-1
    sos.var['type'][var] = Type
    sos.var['Z'][var] = csr_matrix(Z)

    if Type == 'sos':
        T, ZZ = getconstraint(Z)
        sos.var['ZZ'][var] = ZZ
        sos.var['T'][var] = T.T
        sos.var['idx'][var+1] = sos.var['idx'][var] + (Z.shape[0])**2

    elif Type == 'poly':
        sos.var['ZZ'][var] = csr_matrix(Z)
        sos.var['T'][var] = eye(Z.shape[0])
        sos.var['idx'][var+1] = sos.var['idx'][var] + Z.shape[0]

    # Modify existing equations
    for i in range(sos.expr['num']):
        sos.expr['At'][i] = vstack([sos.expr['At'][i], csr_matrix((sos.var['T'][var].shape[0], sos.expr['At'][i].shape[1]))])\
    
    # Modify existing objective
    if isinstance(sos.objective, list):
        sos.objective = csr_matrix((sos.var['idx'][var+1]-sos.var['idx'][var], 1))
    else:
        sos.objective = vstack([sos.objective, csr_matrix((sos.var['idx'][var+1]-sos.var['idx'][var], 1))])

    # Modify decision variable table
    varidx = sos.var['idx'][var+1]-sos.var['idx'][0]
    decvartabletempxxx = symbols(f'coeff_:{varidx}')    # coeff_0, coeff_1, ...
    decvartabletempxxx = decvartabletempxxx[sos.var['idx'][var]-sos.var['idx'][0]:] # only extract necessary coefficients

    for i in decvartabletempxxx:
        sos.symdecvartable.append(i)    # sympy symbols list

    sos.decvartable = str(sos.symdecvartable)   # sympy symbols list to string

    if Type == 'sos':
        if sos.var['ZZ'][var].shape[0] <= 1:
            vartable = sos.symvartable  # sympy symbols list
            temp_ZZ = sos.var['ZZ'][var].toarray()
            ZZSym = []
            for row in temp_ZZ:
                temp_res = [var**deg for var, deg in zip(vartable, row)]
                ZZSym.append(prod(temp_res))
            coeff = myAdecvarprod(sos.var['T'][var], sos.symdecvartable[sos.var['idx'][var]:sos.var['idx'][var+1]])
            V = coeff.T * Matrix(ZZSym)
        else:
            dummy = sos.symdecvartable[sos.var['idx'][var]:sos.var['idx'][var+1]]
            V = Matrix(Zsym).T * Matrix(int(np.sqrt(len(dummy))),int(np.sqrt(len(dummy))),dummy).T * Matrix(Zsym)   # Python reshape and MATLAB reshape is different
            V = V[0,0]
        
    elif Type == 'poly':
        temp_MA = sos.symdecvartable[sos.var['idx'][var]:sos.var['idx'][var+1]]
        V = Matrix(temp_MA).dot(Matrix(Zsym))

    return sos, V




    
    

    


from sympy import sympify
from .sosvar import sosvar

def sossosvar(sos, Zsym):
    '''
    SOSSOSVAR --- Declare a new sum of squares variable in
        an SOS program

    sos, V = sossosvar(SOSP,Z)

    OUTPUTS:
    SOS: the modified sum of squares program structure.
    V: the new sum of squares decision variable. If SOS contains symbolic variables, the
    output type of P will be sym. Otherwise, the output type defaults to
    dpvar unless the caller specifies PVoption='pvar' as a fourth input.

    INPUTS
    SOS: The sum of squares program structure to be modified.
    ZSym[LIST]: the list of monomials contained in the Gram
    decomposition of VAR, i.e.,

        VAR = ZSym' * COEFF * ZSym

    where COEFF is a coefficient matrix that is restricted to be
    positive semidefinite. COEFF and the decision variables contained
    in it will be constructed automatically by SOSSOSVAR.
    '''
    if isinstance(Zsym, int) or Zsym == 1:
        Zsym = sympify(Zsym)

    sos, V = sosvar(sos, 'sos', Zsym)

    return sos, V
    